// ################################################################################
//
//  Reach Robot NANO v0.00 (B0)
//
//  Released:  25/10/2022
//
//  Author: TechKnowTone
//
// ################################################################################
/*
    TERMS OF USE: This software is furnished "as is", without technical support, and
    with no warranty, expressed or implied, as to its usefulness for any purpose. In
    no event shall the author or copyright holder be liable for any claim, damages,
    or other liability, whether in an action of contract, tort or otherwise, arising
    from, out of or in connection with the software or the use or other dealings in
    the software.

    Micro: NANO

    This code was written for Coder Dojo Darlington. It controls a reach robot, driven
    by 4 servo motors, using two X/Y analogue joysticks. The joysticks have built-in
    push switches, and there are 3 button switches at the rear of the control box.
    Inside the control box are two forward facing programmable RGB LEDs. The servo
    motors have been calibrated using a Windows app, via the serial port. This code
    should normally be run in Calibrated = true mode.

    When changing a servo or calibrating a new robot, set Calibrated = false, and
    re-compile the code. Once you have performed the calibration tasks and entered
    the new PWM limits, then set Calibrated = true, for normal use.

    The purpose of the robot is to allow Ninjas to demonstrate there coding skills by
    utilising all of the features of the robots hardware.
*/
// Declare and load external libraries
#include <Servo.h>    // servo library
#include <FastLED.h>  // a fast WS2812B library

// configuration
String Version = "v0.00 (R0)";
String Release = "25/10/2022";

// Define servo calibration constants. These PWM values are determined by following
// the calibration procedure found on my web site. All servos are different, hence
// yuou will need to determine specific values for your robot, and place them here.
#define fwdArmMax 2162    // forward arm Max servo value, default = 1450
#define fwdArmMin 638     // forward arm Min servo value, default = 900
#define fwdArmVert 1205   // forward arm vertical servo value, default = 1100
#define gripClose 864     // jaws closed servo value, default = 1000
#define gripOpen 1100     // jaws moderately open value (23%), default = 1500
#define gripWide 2000     // jaws wide open values, default = 2000
#define turntableCtr 1562 // turntable servo centre value, default = 1500
#define turntableMax 2179 // turntable left 60N servo Max value, default = 1800
#define turntableMin 889  // turntable right 60P servo Min value, default = 1200
#define vertArmMaxA 1612  // vertical arm Max 'A' servo value, default = 1900
#define vertArmMaxB 2079  // vertical arm Max 'B' servo value, default = 2000
#define vertArmMinA 710   // vertical arm Min 'A' servo value, default = 1650
#define vertArmMinB 916   // vertical arm Min 'B' servo value, default = 1650
#define vertArmMinC 1524  // vertical arm Min 'C' servo value, default = 1900
#define servoMinPWM  500  // minimum servo limit when uncalibrated
#define servoMaxPWM 2400  // maximum servo limit when uncalibrated

// calculate fixed differential values
long fwdArmFwdDiff = fwdArmMax - fwdArmVert;
long fwdArmBwdDiff = fwdArmVert - fwdArmMin;
long vertArmFwdMaxDiff = vertArmMaxB - vertArmMaxA;
long vertArmFwdMinDiff = vertArmMinB - vertArmMinA;
long vertArmBwdMaxDiff = vertArmMaxB - vertArmMaxB; // 0
long vertArmBwdMinDiff = vertArmMinC - vertArmMinB;

#define Floor0 turntableCtr   // floor position for servo 0
#define Floor1 1920           // floor position for servo 1
#define Floor2 950            // floor position for servo 2
#define Floor3 gripOpen       // floor position for servo 3
#define Home0 turntableCtr    // home position for servo 0
#define Home1 fwdArmVert      // home position for servo 1
#define Home2 1734            // home position for servo 2
#define Home3 gripClose       // home position for servo 3
#define PWR_tMax 3000         // 3 second time-out for auto-power off
#define PWR_tPing 500         // 500ms time-out on receipt of a ping character
#define Reset0 turntableCtr   // RESET position for servo 0
#define Reset1 fwdArmMin      // RESET position for servo 1
#define Reset2 vertArmMinC    // RESET position for servo 2
#define Reset3 gripClose      // RESET position for servo 3
#define servoOffMax 0         // sets maximum thermal drift offset for servo 0
#define servoOffRmpDwn 100000 // sets thermal offset ramp down time in miliseconds
#define servoOffRmpUp 10000   // sets thermal offset ramps up time in miliseconds

// Define general constants and I/O pins
#define JoyLXpin A1           // pin used for left joystick VRy
#define JoyLYpin A0           // pin used for left joystick VRx
#define JoyRXpin A4           // pin used for right joystick VRy
#define JoyRYpin A5           // pin used for rightt joystick VRx
#define LED_Pin 2             // ESP32 LED GPIO output pin
#define NumLEDs 2             // define the number of RGB LEDs in the chain
#define sw0Pin  3             // left switch SW0 assigned to GPIO13
#define sw1Pin  4             // right switch SW1 assigned to GPIO15
#define sw2Pin  5             // right switch SW1 assigned to GPIO15
#define swJoyL  A2            // left joystick push switch
#define swJoyR  A3            // right joystick push switch

// create a FastLED instance and define the total number of LEDs in the chain
CRGB LED[NumLEDs];

// define arrays
int servoCtr[] = {Reset0,Reset1,Reset2,Reset3}; // value sent to servos
bool servoAtt[4] = {false,false,false,false};   // servo attached flags
int servoMax[] = {turntableMax,fwdArmMax,vertArmMaxB,gripWide};  // servo max pulse
int servoMin[] = {turntableMin, fwdArmMin, vertArmMinA, gripClose};  // servo max pulse
int servoPins[] = {  8,   9,  12,  13 };        // servo output pins assigned
int servoTgt[4];                                // target values for servos
int servoVal[] = {Reset0,Reset1,Reset2,Reset3}; // value sent to servos

// Declare objects
Servo servoInst[4]; // define Reach Robot servos

// Declare and initialise global variables
int Angle;                    // start with servo in the centre pos
int anyInt = 0;               // any temporary variable
bool Calibrated;              // limits functionality, set to true once calibrated
char cmdMode;                 // command mode
int cmdRec;                   // > 0 if a '~' has been received
int cmdSgn;                   // +/- 1 to handle negative numbers, default = 1
char cmdType;                 // command type
int cmdVal;                   // value associated with a cmdType
int incS0;                    // servo[0] increment
int incS1;                    // servo[1] increment
int incS2;                    // servo[2] increment
int incS3;                    // servo[3] increment
int incOffset;                // > 0 flag which controls thermal offset increments
int joyLX;                    // read value of left joystick X
int joyLY;                    // read value of left joystick Y
int joyRX;                    // read value of right joystick X
int joyRY;                    // read value of right joystick Y
char keyChar;                 // any keyboard character
int keyPnt;                   // pointer to last servo affected by keyboard demand
int keyRec;                   // > 0 if a key has been received
int keyVal;                   // any keyboard value
int LED_Cnt;                  // general counter used in LED tasks
int LED_Del;                  // LED Task delay, adds pauses into the processes
int LED_Inc;                  // general increment used in LED tasks
int LED_Mode;                 // LED task pointer
int MainTask;                 // main task pointer
bool moveCalc;                // set to true after servo values have been updated
int moveCnt;                  // counter used in moving to target values
int moveInc;                  // temp move value
bool MOVING;                  // == true when servoces are being adjusted
unsigned long moveInterval;   // move engine interval in microseconds
unsigned long moveMicros;     // move engine interval timer in microseconds
unsigned long next10ms;       // 10ms timer
unsigned long next20ms;       // 20ms timer
bool Once;                    // if = true then coming out of reset
int PWR_timeout;              // disables servos after a time-out, when no code received
int servo2Mod;                // auto-modified servoVal[2]
bool servoEn;                 // global servo attached status
int servoLL;                  // servo min time setting
int servoNum;                 // target servo number 0 - 3 for setting LL/UL limits, default = 0
int servoOff0;                // thermal offset for servo 0
long servoOffDec;             // thermal offset decrement counter
long servoOffDecT;            // thermal offset decrement counter total count
long servoOffInc;             // thermal offset increment counter
long servoOffIncT;            // thermal offset increment counter total count
int servoPin;                 // servo output pin
int servoUL;                  // servo max time setting
int SubDel;                   // subtask delay counter
int SubTask;                  // main task subtask pointer
int sw0Cnt;                   // button switch counter
int sw0DwnTime;               // button pressed down time
bool sw0LastState;            // previous state of button switch, HIGH/LOW
bool sw0State;                // state of read button switch pin
int sw0Timer;                 // timer used to detemine button sequences
int sw1Cnt;                   // button switch counter
int sw1DwnTime;               // button pressed down time
bool sw1LastState;            // previous state of button switch, HIGH/LOW
bool sw1State;                // state of read button switch pin
int sw1Timer;                 // timer used to detemine button sequences
int sw2Cnt;                   // button switch counter
int sw2DwnTime;               // button pressed down time
bool sw2LastState;            // previous state of button switch, HIGH/LOW
bool sw2State;                // state of read button switch pin
int sw2Timer;                 // timer used to detemine button sequences
int swJLCnt;                  // button switch counter
int swJLDwnTime;              // button pressed down time
bool swJLLastState;           // previous state of button switch, HIGH/LOW
bool swJLState;               // state of read button switch pin
int swJLTimer;                // timer used to detemine button sequences
int swJRCnt;                  // button switch counter
int swJRDwnTime;              // button pressed down time
bool swJRLastState;           // previous state of button switch, HIGH/LOW
bool swJRState;               // state of read button switch pin
int swJRTimer;                // timer used to detemine button sequences
int Task10ms;                 // 10ms subtask timer
int Task20ms;                 // 20ms subtask timer
bool ValChg;                  // temporary flag set when a servo value is changed

// ---------------------------------------------------------------------------------

void setup() {
  // put your setup code here, to run once:
  // define switch I/O pins
  pinMode(sw0Pin,INPUT_PULLUP); // button switch sw0 pin
  pinMode(sw1Pin,INPUT_PULLUP); // button switch sw1 pin
  pinMode(sw2Pin,INPUT_PULLUP); // button switch sw1 pin
  pinMode(swJoyL,INPUT_PULLUP); // left joystick switch pin
  pinMode(swJoyR,INPUT_PULLUP); // left joystick switch pin
  
  Serial.begin(115200);         // high baud rate for sweeping functions
  Serial.print(F("\n\n\n\n\n\n\n\n\n\n"));

  // initialise the FastLED component
  FastLED.addLeds<WS2812B, LED_Pin, GRB>(LED, NumLEDs);
  FastLED.clear();              // clear the RGB data array LED[0]...LED[2]
  FastLED.setBrightness(255);   // set the global brightness to 32, max 255
  FastLED.show();               // send the cleared pixel data to the RGB LEDs

  setDefaults();                // reset variables
    
  runPOST();
  
  SyncTimers();
}

// ---------------------------------------------------------------------------------

void loop() {
  //
  // the first part of this loop focuses on running the two movement engines
  // and the associated servo move to target values routine
  // the movement is controlled by a separate loop timer than can be speed adjusted

  
  //###############################################################################
  //
  //  Move calculations
  //
  //###############################################################################
  if (moveCalc) {
    // this code is triggered by the move loop timer, buy only after the servo
    // values have been written, which is synchronise with the loop timer
    moveCalc = false; // reset the flag for the next cycle
    
    // move towards target values if moveCnt > 0
    if (moveCnt > 0) {
      MOVING = true;
      if (moveCnt > 1) {
        // in the middle of a move towards the target value(s)
        for (int zP = 0; zP < 4; zP++) {
          moveInc = (servoTgt[zP] - servoVal[zP])/moveCnt;
          servoVal[zP] += moveInc;
          if (servoVal[zP] != servoTgt[zP]) {ValChg = true;}
        }
      } else {
        // moveCnt == 1 so move to target values
        servoVal[0] = servoTgt[0]; servoVal[1] = servoTgt[1];
        servoVal[2] = servoTgt[2]; servoVal[3] = servoTgt[3];
      } 
      ValChg = true; // trigger a servo update
      if (servoVal[0] != servoTgt[0]) {incOffset = 10;} // increment thermal counter for servo0
      setVertMinMax(); // ensure vertical channel is within limits
      servo2Mod = servoVal[2];
      // decrement the move counter whilst 
      moveCnt--;
    } else {MOVING = false;}  // clear the movement flag once counter == 0
  }  

  //###############################################################################
  //
  //  10ms Move Servos if ValChg == true
  //
  //###############################################################################
  if ((micros() - moveMicros) >= moveInterval) {
    // normally runs every 10ms, but can be varied to suit speed of
    // robot movements
    moveMicros = micros();
        
    // if any servo value has changed we update all servos at this point
    if (ValChg) {
      ValChg = false;
      // one or more servo values have changed, so update all of them
      if (!servoEn) {attachServos(0);}
      servoInst[0].writeMicroseconds(servoVal[0]);
      servoInst[1].writeMicroseconds(servoVal[1]);
      servoInst[2].writeMicroseconds(servo2Mod);
      servoInst[3].writeMicroseconds(servoVal[3]);
    }
    moveCalc = true; // set the flag to run the calculations in the next cycle
  }

  
  //###############################################################################
  //
  //  Read serial port
  //
  //###############################################################################
  // read keyboard input when not doing other things
  readSerial(); // empty Rx buffer
  if (keyVal != -1) {keyRec = 10;} // set a timer-out for adjustment speed

  
  //###############################################################################
  //
  //  10ms Tasks
  //
  //###############################################################################
  if (Task10ms > 0) {
    switch(Task10ms) {
      case 1: readSwJR(); break;
      case 2: readSwJL(); break;
      case 3: readSW2(); break;
      case 4: readSW1(); break;
    } Task10ms--; // count down the sub-tasks from 3 - 1 as we loop
  } else if ((millis() - next10ms) >= 10) {
    // do these every 10ms
    next10ms = millis(); // set next time point
    Task10ms = 4;   // initiate 10ms task sequence
    readSW0();      // read left button switch SW0 as priority task
  }

  
  //###############################################################################
  //
  //  20ms Tasks
  //
  //###############################################################################
  if (Task20ms > 0) {
    switch(Task20ms) {
      case 1: LED_Tasks(); break;     // various LED tasks
      case 2: JoyCtrl(); break;       // respond to joystick demands
      case 3: readJoyRY(); break;     // read right joystick Y
      case 4: readJoyRX(); break;     // read right joystick X
      case 5: readJoyLY(); break;     // read left joystick Y
      case 6: readJoyLX(); break;     // read left joystick X
    } Task20ms--; // count down the sub-tasks from 3 - 1 as we loop
  } else if ((millis() - next20ms) >= 20) {
    // do these every 20ms
    next20ms = millis(); // set next time point
    Task20ms = 6;

    // perform main tasks
    switch (MainTask) {
      case 1: MoveToRest(); break;
      case 2: MoveToHome(turntableCtr,gripClose); break;
      case 3: MoveToFloor(); break;
      case 4: MoveToHome(turntableMax,gripOpen); break;
      case 5: MoveToHome(turntableMin,gripOpen); break;
    }
  }

  // respond to remote controller pings
  if (cmdRec) {Serial.println(F("~")); cmdRec = 0;}
}

// ---------------------------------------------------------------------------------

void runPOST() {
  // called during start-up
  Serial.print(F("Reach Robot NANO ")); Serial.println(Version);
  Serial.print(F("Release: ")); Serial.println(Release);
  Serial.println(F("Starting POST..."));
  
  if (Calibrated) {
    Serial.println(F("Centring Servos..."));
    // set LEDs to red warning for 3 seconds
    for (int zT = 0; zT < 15; zT++) {LED_Warning(); delay(200);}
    attachServos(100); delay(200); detachServos();
  }

  // all done
  Serial.println(F("POST complete!"));
}

// ---------------------------------------------------------------------------------

void setDefaults() {
  // load default values, at start and after a RESET
  Calibrated = true;      // limits functionality, set to true once calibrated
  cmdMode = ' ';          // command mode
  cmdRec = 0;             // > 0 if a '~' or any data has been received
  cmdSgn = 1;             // +/- 1 to handle negative numbers, default = 1
  cmdType = ' ';          // command type
  cmdVal = 0;             // value associated with a cmdType
  incOffset = 0;          // > 0 flag which controls thermal offset increments
  joyLX = 512;            // read value of left joystick X
  joyLY = 512;            // read value of left joystick Y
  joyRX = 512;            // read value of right joystick X
  joyRY = 512;            // read value of right joystick Y
  keyPnt = -1;            // pointer to last servo affected by keyboard demand
  keyVal = -1;            // any keyboard value
  LED_Cnt = 0;            // general counter used in LED tasks
  LED_Del = 0;            // LED Task delay, adds pauses into the processes
  LED_Inc = 1;            // general increment used in LED tasks
  LED_Mode = 1;           // LED task pointer
  MainTask = 0;           // main task pointer
  moveCalc = false;       // set to true after servo values have been updated
  MOVING = false;         // == true when servoces are being adjusted
  moveInterval = 10000;   // move engine interval nominally 10ms
  Once = true;            // if = true then coming out of reset
  PWR_timeout = 0;        // disables servos after a time-out, when no code received
  servoLL = 500;          // servo min PWM setting
  servoNum = 0;           // target servo number 0 - 3 for setting LL/UL limits, default = 0
  servoOff0 = 0;          // offset for servo 0
  servoOffDec = 10000;    // offset decrement counter
  servoOffInc = 0;        // offset increment counter
  servoPin = -1;          // servo output pin is undefined
  servoUL = 2400;         // servo max PWM setting

  setCalibrated(Calibrated);  // set limits to match calibrated state

  SubDel = 0;             // SubTask delay counter
  SubTask = 0;            // main task subtask pointer
  sw0Cnt = 0;             // button switch counter
  sw0DwnTime = 0;         // button pressed down time
  sw0LastState = HIGH;    // previous state of button switch, HIGH/LOW
  sw0State = HIGH;        // state of read button switch pin
  sw0Timer = 0;           // timer used to detemine button sequences
  sw1Cnt = 0;             // button switch counter
  sw1LastState = HIGH;    // previous state of button switch, HIGH/LOW
  sw1State = HIGH;        // state of read button switch pin
  sw1Timer = 0;           // timer used to detemine button sequences
  sw2Cnt = 0;             // button switch counter
  sw2LastState = HIGH;    // previous state of button switch, HIGH/LOW
  sw2State = HIGH;        // state of read button switch pin
  sw2Timer = 0;           // timer used to detemine button sequences
  swJLCnt = 0;            // button switch counter
  swJLDwnTime = 0;        // button pressed down time
  swJLLastState = HIGH;   // previous state of button switch, HIGH/LOW
  swJLState = HIGH;       // state of read button switch pin
  swJLTimer = 0;          // timer used to detemine button sequences
  swJRCnt = 0;            // button switch counter
  swJRDwnTime = 0;        // button pressed down time
  swJRLastState = HIGH;   // previous state of button switch, HIGH/LOW
  swJRState = HIGH;       // state of read button switch pin
  swJRTimer = 0;          // timer used to detemine button sequences
  Task10ms = 0;           // 10ms subtask timer
  Task20ms = 0;           // 20ms subtask timer
  ValChg = false;         // flag set when a servo value is changed
}

// ----------------------------------------------------------------------

void SyncTimers() {
  // synchronises timers
  // offsets are used to spacer out timer actions
  while (millis() < 20) {}
  
  moveMicros = micros() - 20000;
  next10ms = millis();
  next20ms = millis() - 5;
}

// ---------------------------------------------------------------------------------
